this.name           = "NShields";
this.author         = "Ngalo";
this.copyright      = "(C) 2015 'Ngalo'.";
this.licence        = "CC-NC-by-SA 3.0";
this.description    = "Main script for N-Shields NPC shield simulation and visual effects";

//=============================================================================================================================
//Predefined flasher colour lists (must be defined before settings). Please do not change but feel free to copy/paste and modify or add your own here.
this._multicolour = ["1 0 0 0.6", "1 0.5 0 0.6", "1 1 0 0.7", "0.3 1 0 0.8", "0 0.5 0.5 0.8", "0 0.5 0.5 0.8", "0 0 1 0.9", "0 0 1 0.9", "0.5 0 0.5 1"];
this._green = ["0 0.7 0 0.7", "0 0.8 0 0.8", "0 0.9 0 0.9", "0 1 0 1"];
this._blue = ["0 0 0.5 0.6", "0 0 0.6 0.7", "0 0 0.7 0.8", "0 0 0.8 0.9", "0 0 0.9 1", "0 0 0.9 1", "0 0 1 1", "0 0 1 1", "0.5 0 1 1"];

//Exploder role lists; size is added automatically later
this._expl_starWhite = ["N-Shields-explode-star-white"];
this._expl_starMulticolour = ["N-Shields-explode-star-red", "N-Shields-explode-star-red", "N-Shields-explode-star-white", "N-Shields-explode-star-green", "N-Shields-explode-star-green", "N-Shields-explode-star-blue"];
this._expl_starGreen = ["N-Shields-explode-star-green"];
this._expl_starBlue = ["N-Shields-explode-star-blue"];
this._expl_ringGreen = ["N-Shields-explode-ring-green"];

//=============================================================================================================================
// SETTINGS

this.$npcShieldConfig = "fore/aft 128:2.0";
this.$thargoidShieldConfig = "dorsal/ventral 192:1.5"; //was "single 128:1.5"; in earlier versions
this.$thargoidEnergyReduction = 192; // so shields don't make them too tough

this.$npcShieldColours = this._multicolour;
this.$npcShieldExploders = this._expl_starMulticolour;
this.$thargoidShieldColours = this._green;
this.$thargoidShieldExploders = this._expl_ringGreen;

this.$drawFlashers = true;
this.$drawExploders = true;
this.$drawHemiGlow = false;
this.$drawSparkEffects = true;

this.$flasherScale = 2;
this.$shield_charge_energybank_threshold = 0.0; // imitates the corresponding global-settings.plist property for ships with fore/aft shields

//=============================================================================================================================
//Settings for Library OXP in-game config
this._NShields_libSettings = {
	Name: this.name,
	Display: "N-Shields",
	Alive: "_NShields_libSettings",
	Bool: {
		B0: {
			Name: "$drawFlashers",
			Def: true,
			Desc: "flasher effects on/off"
		},
		B1: {
			Name: "$drawExploders",
			Def: true,
			Desc: "star-explosion effects on/off"
		},
		B2: {
			Name: "$drawHemiGlow",
			Def: false,
			Desc: "hemisphere glow effects on/off"
		},
		B3: {
			Name: "$drawSparkEffects",
			Def: true,
			Desc: "unshielded hit spark-shower effects on/off"
		}
	},
	SInt: {
		S0: {
			Name: "$flasherScale",
			Def: 2,
			Min: 0.5,
			Max: 5,
			Desc: "Scale factor for flashers, relative to collision radius",
			Float: true
		}
	}
}

// Array for other NPC shipTakingDamage-like handlers
this._damageHandlers = []; //imitates player shipTakingDamage, called for all hits but amount=0 if shields still up

this.startUp = function()
{
	//Shield config dictionary
	this.$shieldTypes = 
	{
		"single": [this.$configureSingleShield, 1],
		"fore/aft": [this.$configureForeAftShields, 2],
		"dorsal/ventral": [this.$configureDorsalVentralShields, 2],
		"fake": [this.$configureFakeShield, 0]
	}
	this.$thargoidshieldArray = this.$thargoidShieldConfig.split(" ")
	this.$npcshieldArray = this.$npcShieldConfig.split(" ")
	
	if (worldScripts.NPC_Equipment_Damage) worldScripts.NPC_Equipment_Damage.NPC_equipmentDamagedHandlers.push(this.NPC_shield_equipmentDamaged);
}
this.startUpComplete = function()
{
	if (worldScripts.Lib_Config) worldScripts.Lib_Config._registerSet(this._NShields_libSettings);
}

this.shipSpawned = function (ship)
{
	if (!this.$shipHasShields(ship)) {return;}

	if (!ship.script.NShields_damageHandlers) // may have been set up already by other OXPs
	{
		ship.script.NShields_damageHandlers = this._damageHandlers.slice(); // need to explicitly copy, not just reference
	}
	if (ship.script.shipTakingDamage)
	{
		ship.script.NShields_damageHandlers.push(ship.script.shipTakingDamage); //store existing handler, if any
	}
	
	if (ship.isFrangible) this.$configureSubEntities(ship);

	if (this.$configureShieldsFromScriptInfo(ship)) return; //try to set up shields from script_info if defined
	//if (this.$configureShieldsFromShipdata(ship)) return;
	if (ship.isThargoid)
	{
		if (this.$configureDefaultThargoidShields(ship)) return;
	}
	this.$configureDefaultNPCShields(ship)
}

//============= Shield configuration functions (stage 1: identify shield type) ============================================================

this.$shipHasShields = function (ship)
{
	if (ship.scriptInfo && ship.scriptInfo.npc_shields == "no") {return false};
	return (ship.isPiloted && !(ship.primaryRole == "escape-capsule") && !ship.isStation) // isPiloted rules out buoys, rocks, weapons and debris
}

this.$configureSubEntities = function(ship)
{
	if (ship.scriptInfo && ship.scriptInfo.NShields_protect_subentities == "no")
	{
		for (sE in ship.subEntities)
		{
			this.$configureShieldsFromScriptInfo(ship.subEntities[sE]) //give subent its own shields if defined, otherwise leave it
		}
		return;
	}
	//otherwise try to protect the subEntity with its owner's shields:
	for (sE in ship.subEntities)
	{
		if (ship.subEntities[sE].script.shipTakingDamage)
		{
			ship.subEntities[sE].script.NShields_hold_shipTakingDamage = ship.subEntities[sE].script.shipTakingDamage;
		}
		ship.subEntities[sE].script.shipTakingDamage = this.NShields_subEntity_override_shipTakingDamage;
	}
}

this.$configureShieldsFromScriptInfo = function(ship)
{
	if (ship.scriptInfo && ship.scriptInfo.NShields_config)
	{
		var config = ship.scriptInfo.NShields_config.split(" ");
		if(this.$configureShieldsFromArray(ship, config)) {return true;}
	}
	return false;
}

/*this.$configureShieldsFromShipdata = function(ship) // TODO - or not as shipdata keys for shields seem not to exist
{
	//var sd = Ship.shipDataForKey(ship.dataKey)
	return false;
}*/

this.$configureDefaultThargoidShields = function(ship)
{
	if (this.$configureShieldsFromArray(ship, this.$thargoidshieldArray)) return true;
	log(this.name, "Thargoid shield config failed. Have you set an invalid default?")
	return false;
}

this.$configureDefaultNPCShields = function(ship)
{
	if (this.$configureShieldsFromArray(ship, this.$npcshieldArray)) return true;
	log(this.name, "standard NPC shield config failed, have you set an invalid default?")
	return false;
}

this.$configureShieldsFromArray = function(ship, configArray)
{
	var type = this.$shieldTypes[configArray[0]];
	if (type)
	{
		var args = [ship];
		args = args.concat(configArray.slice(1, type[1] + 1));
		type[0].apply(this, args)
	}
	else return false; // invalid shield type identifier
	
	//if we got here without returning, the shields are configured OK so apply the other stuff
	this.$applyEnergyReduction(ship);
	this.$configureShieldColours(ship);
	this.$configureShieldExploders(ship);
	return true;
}

//========== Shield configuration stage 2: actually apply stuff to the ship ================================

this.$configureSingleShield = function(ship, shieldData1)
{
	log (this.name, "Single shieldData: "+shieldData1)
	var labels = ship.script.NShields_labels = {shields: ["NShields_shieldLevel"],
						    maxShields: ["NShields_maxShield"],
						    rechargeRates: ["NShields_shieldRechargeRate"]
						    };
	
	ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.NShields_Single_override_shipTakingDamage;
	ship.script.NShields_updateRecharge = this.$updateRecharge;
	
	this.$configureShields(ship, labels, [shieldData1])
	this.$applyShieldUpgrades(ship, labels)
	this.$activateShields(ship, labels)
}

this.$configureForeAftShields = function(ship, shieldData1, shieldData2)
{
	var labels = ship.script.NShields_labels = {shields: ["forwardShield", "aftShield"],
						    maxShields: ["maxForwardShield", "maxAftShield"],
						    rechargeRates: ["forwardShieldRechargeRate", "aftShieldRechargeRate"]
						    };
	
	ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.NShields_ForeAft_override_shipTakingDamage;
	ship.script.NShields_updateRecharge = this.$updateRecharge_ForeAft;
	
	this.$configureShields(ship, labels, [shieldData1, shieldData2])
	this.$applyShieldUpgrades(ship, labels)
	this.$activateShields(ship, labels)
}

this.$configureDorsalVentralShields = function(ship, shieldData1, shieldData2)
{
	var labels = ship.script.NShields_labels = {shields: ["dorsalShield", "ventralShield"],
						    maxShields: ["maxDorsalShield", "maxVentralShield"],
						    rechargeRates: ["dorsalShieldRechargeRate", "ventralShieldRechargeRate"]
						    };
	ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.NShields_DorsalVentral_override_shipTakingDamage;
	ship.script.NShields_updateRecharge = this.$updateRecharge;
	
	this.$configureShields(ship, labels, [shieldData1, shieldData2])
	this.$applyShieldUpgrades(ship, labels)
	this.$activateShields(ship, labels)
}

this.$configureFakeShield = function(ship)
{
	ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.NShields_Fake_override_shipTakingDamage;
}

this.$configureShields = function(ship, shieldLabels, shieldData) //requires all shieldData in one array, not separate
{
	for (i in shieldLabels.maxShields)
	{
		if (shieldData[i])
		{
			var sd = shieldData[i].split(":");
		}
		else //use last shield config in list, if fewer than necessary for the shield type are defined
		{
			for (i in shieldData.reverse())
			{
				if (shieldData[i]) var sd = shieldData[i].split(":");
			}
		}
		if (sd == undefined) log(this.name, "ship "+ship.name+" had shield config failure: "+shieldData)
		
		ship[shieldLabels.maxShields[i]] = parseInt(sd[0]);
		ship[shieldLabels.rechargeRates[i]] = parseFloat(sd[1]);
	}
	ship.script.NShields_lastRechargeTime = 0;
	ship.script.NShields_rechargeTimer = new Timer(ship.script, ship.script.NShields_updateRecharge, -1, 1);
}

this.$applyShieldUpgrades = function(ship, labels) //TODO- other upgrades
{
	//log ("applying shield upgrades to "+ ship.name) 
	if (ship.equipmentStatus("EQ_SHIELD_BOOSTER") == "EQUIPMENT_OK")
	{
		let eq = EquipmentInfo.infoForKey("EQ_SHIELD_BOOSTER");
		let shield_boost = parseInt(eq.scriptInfo.oolite_shield_increase);
		let recharge_boost = parseFloat(eq.scriptInfo.oolite_shield_recharge_multiplier);
		for (i in labels.maxShields)
		{
			ship[labels.maxShields[i]] += shield_boost;
			if (recharge_boost) ship[labels.rechargeRates[i]] *= recharge_boost;
		}
		ship.removeEquipment("EQ_SHIELD_BOOSTER")
		ship.awardEquipment("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") // dummy equipment for shipconfig compatibility
	}
	if (ship.equipmentStatus("EQ_SHIELD_ENHANCER") == "EQUIPMENT_OK")
	{
		let eq = EquipmentInfo.infoForKey("EQ_NAVAL_SHIELD_BOOSTER");
		let shield_boost = parseInt(eq.scriptInfo.oolite_shield_increase);
		let recharge_boost = parseFloat(eq.scriptInfo.oolite_shield_recharge_multiplier);
		for (i in labels.maxShields)
		{
			ship[labels.maxShields[i]] += shield_boost;
			if (recharge_boost) ship[labels.rechargeRates[i]] *= recharge_boost;
		}
		ship.removeEquipment("EQ_SHIELD_ENHANCER")
		ship.awardEquipment("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER")
	}
	
	//Shield Cycler, if that OXP is installed (Work In Progress):
	/*
	if (worldScripts["SC Core"])
	{
		if (Math.random() < 0.1) //TODO: proper allocation
		{
			ship.awardEquipment("EQ_SHIELD_CYCLER_INTERNAL") //TODO: variants?
			if (ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") == "EQUIPMENT_OK")
			{
				//Advanced version requires Military Shield Enhancer
				ship.script.NShields_SC_version = 3;
				ship.script.NShields_SC_manual_version = 0; //For now, NPCs always use automatic mode //Math.round(2 + Math.random()); // 2 or 3 = Standard or Advanced Manual Configurator 
			}
			else if (ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") == "EQUIPMENT_OK")
			{
				//Standard version, requires Shield Booster
				ship.script.NShields_SC_version = 2;
				ship.script.NShields_SC_manual_version = 0;
			}
			else
			{
				//Basic version
				ship.script.NShields_SC_version = 1;
				ship.script.NShields_SC_manual_version = 0;
			}
			
			ship.script.NShields_damageHandlers.push(this.NPC_sc_adjust) // ?
		}
	}
	*/
	
	// ShipConfiguration OXP shield upgrades/downgrades
	if (worldScripts.ShipConfiguration_Core && ship.maxAftShield != undefined && ship.maxForwardShield != undefined) //ensure the ship actually has fore/aft shields
	{
		let shipconf_fwdShield = worldScripts.ShipConfiguration_Core.$equipmentItemInUse("frontshields", ship);
		let shipconf_aftShield = worldScripts.ShipConfiguration_Core.$equipmentItemInUse("aftshields", ship);
		
		if (shipconf_fwdShield && shipconf_fwdShield != ship.script.NShields_shipconf_fwdShield) //in case we ran this function already
		{
			if (shipconf_fwdShield == "EQ_NO_FORWARD_SHIELD") ship.maxForwardShield = 0;
			else ship.maxForwardShield += parseInt(EquipmentInfo.infoForKey(shipconf_fwdShield).scriptInfo.oolite_shield_increase);
			ship.script.NShields_shipconf_fwdShield = shipconf_fwdShield;
		}
		
		if (shipconf_aftShield && shipconf_aftShield != ship.script.NShields_shipconf_aftShield)
		{
			if (shipconf_aftShield == "EQ_NO_AFT_SHIELD") ship.maxAftShield = 0;
			else ship.maxAftShield += parseInt(EquipmentInfo.infoForKey(shipconf_aftShield).scriptInfo.oolite_shield_increase);
			ship.script.NShields_shipconf_aftShield = shipconf_aftShield;
		}
		//log (this.name, "applied shield upgrades "+shipconf_fwdShield+", "+shipconf_aftShield+" to "+ship.name)
		//log ("fwd "+ship.forwardShield+" aft "+ship.aftShield)
		//TODO: call $updateShipSpec?
	}
}

this.$activateShields = function(ship, shieldLabels)
{
	for (S in shieldLabels.shields)
	{
		ship[shieldLabels.shields[S]] = ship[shieldLabels.maxShields[S]];
	}
}

this.$applyEnergyReduction = function(ship) //optionally used for ships which would be too hard to kill if they had shields and their full usual energy banks
{
	if (ship.scriptInfo && ship.scriptInfo.NShields_energyReduction)
	{
		ship.maxEnergy -= parseInt(ship.scriptInfo.NShields_energyReduction);
		return;
	}
	if (ship.isThargoid)
	{
		ship.maxEnergy -= this.$thargoidEnergyReduction;
	}
}

this.$configureShieldColours = function(ship)
{
	if (ship.scriptInfo && ship.scriptInfo.NShields_colours)
	{
		ship.script.NShields_flareColours = ship.scriptInfo.NShields_colours;
		return;
	}
	if (ship.isThargoid) {ship.script.NShields_flareColours = this.$thargoidShieldColours; return;}
	ship.script.NShields_flareColours = this.$npcShieldColours;
}

this.$configureShieldExploders = function(ship)
{
	if (ship.collisionRadius > 128) ship.script.NShields_sizeClass = "-XL";
	else if (ship.collisionRadius > 64) ship.script.NShields_sizeClass = "-L";
	else if (ship.collisionRadius > 32) ship.script.NShields_sizeClass = "-M";
	else if (ship.collisionRadius > 16) ship.script.NShields_sizeClass = "-S";
	else ship.script.NShields_sizeClass = "-XS";
	
	if (ship.scriptInfo && ship.scriptInfo.NShields_exploders)
	{
		ship.script.NShields_exploders = ship.scriptInfo.NShields_exploders;
		return;
	}
	if (ship.isThargoid) {ship.script.NShields_exploders = this.$thargoidShieldExploders; return;}
	ship.script.NShields_exploders = this.$npcShieldExploders;
}

//=========== Damage handlers ===============================================================================
//=========== For these functions 'this' must mean a ship script ============================================ 

this.NShields_Single_override_shipTakingDamage = function (amount, whom, type, caller)
{
	if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed")
	{
		//cases where, for practical purposes, no shield handling is needed - but other OXPs' handlers may still need to run
		worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
		return;
	}
	this.NShields_updateRecharge.apply(this, [clock.absoluteSeconds - this.NShields_lastRechargeTime]);
	worldScripts.NShields.$resetRechargeTimer.call(this);
	if (caller) caller.energy += amount; //heal subEntity if called from there
	else this.ship.energy += amount;
	
	if (type == "heat damage") // will not be able to locate hit, and different handling is required anyway
	{
		worldScripts.NShields.$applyHeatDamage(this.ship, amount);
		return;
	}
	
	if (this.ship.isVisible) var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom);
	
	this.ship.NShields_shieldLevel -= amount;
	if (this.ship.NShields_shieldLevel < 0)
	{
		if (caller) caller.energy += this.ship.NShields_shieldLevel;
		else this.ship.energy += this.ship.NShields_shieldLevel; // add negative shield value to energy
		worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.NShields_shieldLevel, whom, type);
		if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.NShields_shieldLevel);
		this.ship.NShields_shieldLevel = 0;
	}
	else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
	
	if (this.ship.isVisible)
	{
		worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.NShields_shieldLevel, this.ship, whom)
	}	
}

this.NShields_ForeAft_override_shipTakingDamage = function (amount, whom, type, caller) //'caller' is added when protecting frangible subEntities 
{
	if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed")
	{
		//cases where, for practical purposes, no shield handling is needed - but other OXPs' handlers may still need to run
		worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
		return;
	}
	
	this.NShields_updateRecharge.apply(this, [clock.absoluteSeconds - this.NShields_lastRechargeTime]);
	worldScripts.NShields.$resetRechargeTimer.call(this);
	if (caller) caller.energy += amount; //heal subEntity if called from there
	else this.ship.energy += amount;
	
	if (type == "heat damage") // will not be able to locate hit, and different handling is required anyway
	{
		worldScripts.NShields.$applyHeatDamage(this.ship, amount);
		if (isNaN(this.ship.forwardShield)) log (this.name, "ERROR: Forward shield is NaN (check 0) on ship "+this.ship);
		if (isNaN(this.ship.aftShield)) log (this.name, "ERROR: Aft shield is NaN (check 0) on ship "+this.ship);
		return;
	}
	
	var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom)
	var hitLocationRelative = hitLocation.subtract(this.ship.position);
	if (hitLocationRelative.dot(this.ship.vectorForward) > 0) // forward shield
	{
		this.ship.forwardShield -= amount;
		if (isNaN(this.ship.forwardShield)) log (this.name, "ERROR: Forward shield is NaN (check 1) on ship "+this.ship+"; amount="+amount);
		
		if (this.ship.forwardShield < 0)
		{
			if (caller) caller.energy += this.ship.forwardShield;
			else this.ship.energy += this.ship.forwardShield; // add negative shield value to energy
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.forwardShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.forwardShield);
			this.ship.forwardShield = 0;
			if (isNaN(this.ship.forwardShield)) log (this.name, "ERROR: Forward shield is NaN (check 2) on ship "+this.ship);
		}
		
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.forwardShield, this.ship, whom, "forwardShield")
	}
	else
	{
		this.ship.aftShield -= amount;
		if (isNaN(this.ship.aftShield)) log (this.name, "ERROR: Aft shield is NaN (check 1) on ship "+this.ship);
		if (this.ship.aftShield < 0)
		{
			if (caller) caller.energy += this.ship.aftShield;
			else this.ship.energy += this.ship.aftShield;
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.aftShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.aftShield);
			this.ship.aftShield = 0;
			if (isNaN(this.ship.aftShield)) log (this.name, "ERROR: Aft shield is NaN (check 2) on ship "+this.ship);
		}
		
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.aftShield, this.ship, whom, "aftShield")
	}
	worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);	
}

this.NShields_DorsalVentral_override_shipTakingDamage = function (amount, whom, type, caller) //'caller' is added when protecting frangible subEntities 
{
	if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed")
	{
		//cases where, for practical purposes, no shield handling is needed - but other OXPs' handlers may still need to run
		worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
		return;
	}
	
	this.NShields_updateRecharge.apply(this, [clock.absoluteSeconds - this.NShields_lastRechargeTime]);
	worldScripts.NShields.$resetRechargeTimer.call(this);
	if (caller) caller.energy += amount; //heal subEntity if called from there
	else this.ship.energy += amount;
	
	if (type == "heat damage") // will not be able to locate hit, and different handling is required anyway
	{
		worldScripts.NShields.$applyHeatDamage(this.ship, amount);
		return;
	}
	
	var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom)
	var hitLocationRelative = hitLocation.subtract(this.ship.position);
	if (hitLocationRelative.dot(this.ship.vectorUp) > 0) // dorsal shield
	{
		this.ship.dorsalShield -= amount;
		if (this.ship.dorsalShield < 0)
		{
			if (caller) caller.energy += this.ship.dorsalShield;
			else this.ship.energy += this.ship.dorsalShield; // add negative shield value to energy
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.dorsalShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.dorsalShield);
			this.ship.dorsalShield = 0;
		}
		else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
		
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.dorsalShield, this.ship, whom)
	}
	else
	{
		this.ship.ventralShield -= amount;
		if (this.ship.ventralShield < 0)
		{
			if (caller) caller.energy += this.ship.ventralShield;
			else this.ship.energy += this.ship.ventralShield;
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.ventralShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.ventralShield);
			this.ship.ventralShield = 0;
		}
		else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
		
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.ventralShield, this.ship, whom)
	}	
}

this.NShields_Fake_override_shipTakingDamage = function(amount, whom, type, caller)
{
	var fakeShield = this.ship.energy - Math.min(64, this.ship.energy / 2);
	worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
	if (this.ship.isVisible)
	{
		var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom);
		if (fakeShield > 0) worldScripts.NShields.$drawShieldFX(hitLocation, amount, this.ship, whom);
		else worldScripts.NShields.$drawSparks(hitLocation, this.ship, amount);
	}
}

this.NShields_subEntity_override_shipTakingDamage = function(amount, whom, type)
{
	this.ship.owner.script.NShields_override_shipTakingDamage(amount, whom, type, this.ship) //not simple 'shipTakingDamage' in case another OXP re-overrode my override
	if (this.NShields_hold_shipTakingDamage) this.NShields_hold_shipTakingDamage(amount, whom, type);	
}
//========== Common damage handler components ================================================================

this.$locateHit = function (ship, attacker) //basic hit location procedure, based on customshields- NOTE needs compatibility for 1.83 weapon pos array--done by removing weapon pos elements, for now
{
	if (!attacker)
	{
		if (player.ship.isCloaked && ship.isVisible) attacker = player.ship;
		else if (ship.script.NShields_cloaked_attacker && ship.script.NShields_cloaked_attacker.position.distanceTo(ship.position) < 25600) attacker = ship.script.NShields_cloaked_attacker;
		else attacker = ship.script.NShields_cloaked_attacker = this.$findCloakedAttacker(ship);
		if (!attacker) //still not found anyone?
		{
			log(this.name, "WARNING: attacker is null; could not find cloaked attacker")
			return ship.position;
		}
	}
	
	/* attempt to improve hit location code, does not work well for NPCs?
	var shieldRadius = ship.collisionRadius * 1.2;

	var hitLocation = attacker.position.add(attacker.vectorForward.multiply(attacker.position.distanceTo(ship.position) - shieldRadius)); // possible forward-weapon hit location
	if (ship.position.distanceTo(hitLocation) > shieldRadius)// + impactradiusfactor) // too far from ship, try other options
	{
		hitLocation = attacker.position.add(attacker.vectorForward.multiply(-(attacker.position.distanceTo(ship.position) - shieldRadius))); //next try aft weapon
		if (ship.position.distanceTo(hitLocation) > shieldRadius)// + impactradiusfactor)
		{
			hitLocation = attacker.position.add(attacker.vectorRight.multiply(attacker.position.distanceTo(ship.position) - shieldRadius)); //then starboard
			if (ship.position.distanceTo(hitLocation) > shieldRadius)// + impactradiusfactor)
			{
				hitLocation = attacker.position.add(attacker.vectorRight.multiply(-(attacker.position.distanceTo(ship.position) - shieldRadius))); //then port
				if (ship.position.distanceTo(hitLocation) > shieldRadius)// + impactradiusfactor)
				{
					hitLocation = Vector3D.interpolate(ship.position, attacker.position, shieldRadius / attacker.position.distanceTo(ship.position)) // fallback method, directly between ships
					//log ("used fallback hit location")
				}
				
			}
		}
	}
	*/
	
	var impactradiusfactor = ((3 * (1500 / (ship.position.distanceTo(attacker.position)))) + 16); // copied from customshields, not understood
	var hitLocation = attacker.position.add(attacker.vectorForward.multiply(attacker.position.distanceTo(ship.position) - (ship.collisionRadius + 1))); // possible forward-weapon hit location
	if (ship.position.distanceTo(hitLocation) > ship.collisionRadius + impactradiusfactor) // too far from ship, try other options
	{
		hitLocation = attacker.position.add(attacker.vectorForward.multiply(-(attacker.position.distanceTo(ship.position) - (ship.collisionRadius + 1))));
		if (ship.position.distanceTo(hitLocation) > ship.collisionRadius + impactradiusfactor)
		{
			hitLocation = attacker.position.add(attacker.vectorRight.multiply(attacker.position.distanceTo(ship.position) - (ship.collisionRadius + 1)));
			if (ship.position.distanceTo(hitLocation) > ship.collisionRadius + impactradiusfactor)
			{
				hitLocation = attacker.position.add(attacker.vectorRight.multiply(-(attacker.position.distanceTo(ship.position) - (ship.collisionRadius + 1))));
				if (ship.position.distanceTo(hitLocation) > ship.collisionRadius + impactradiusfactor)
				{
					hitLocation = Vector3D.interpolate(ship.position, attacker.position, ship.collisionRadius / attacker.position.distanceTo(ship.position)) // fallback method, directly between ships
					//log ("used fallback hit location")
				}
				
			}
		}
	}
	return hitLocation;
}

this.$findCloakedAttacker = function(ship)
{
	function $isCloakedAttacker(entity)
	{
		return (entity.isShip && !entity.isPlayer && entity.isCloaked && entity.target);
	}
	var scan = system.filteredEntities(this, $isCloakedAttacker, ship, 25600);
	for (i in scan)
	{
		if (scan[i].target == ship) return scan[i];
	}
}

/*this.$isCloakedAttacker = function(entity)
{
	return entity.isShip && !entity.isPlayer && entity.isCloaked && entity.target);
}*/

//note 'this' variable must point to ship script for both recharge funcs when called
this.$updateRecharge = function(delta) //delta is time since last call (so this could be used as a frame callback, though it isn't)
{
	var fullShields = 0;
	if (delta == undefined) delta = 1; //called by timer so no parameter passed
	if (this.ship == player.ship.target) log (this.name, "updating recharge for "+this.ship.name+" delta: "+ delta);
	this.NShields_lastRechargeTime = clock.absoluteSeconds;
	for (S in this.NShields_labels.shields)
	{
		if (this.ship[this.NShields_labels.shields[S]] < this.ship[this.NShields_labels.maxShields[S]])
		{
			var recharge = Math.min((this.ship[this.NShields_labels.rechargeRates[S]] * delta), (this.ship[this.NShields_labels.maxShields[S]] - this.ship[this.NShields_labels.shields[S]]));
			if (this.ship == player.ship.target) log (this.name, "calculated recharge: "+recharge+" for "+this.ship[this.NShields_labels.shields[S]]);
			
			if (recharge <= this.ship.energy) // check sufficient energy to charge shield
			{
				this.ship[this.NShields_labels.shields[S]] += recharge;
				this.ship.energy -= recharge;
			}
			else if (this.ship == player.ship.target) log(this.name, "insufficient energy "+this.ship.energy);
		}
		else fullShields++; //note shield is full
	}
	if (fullShields == this.NShields_labels.shields.length) //all fully charged?
	{
		this.NShields_rechargeTimer.stop();
	}
}

this.$updateRecharge_ForeAft = function(delta)
{
	//Debugging:
	var debugNeeded = false;
	if (isNaN(this.ship.forwardShield))
	{
		log(this.name, "ERROR: forward shield NaN on ship "+this.ship)
		debugNeeded = true;
		//and attempt fix:
		this.ship.forwardShield = 0;
	}
	if (isNaN(this.ship.aftShield))
	{
		log(this.name, "ERROR: aft shield NaN on ship "+this.ship)
		debugNeeded = true;
		this.ship.aftShield = 0;
	}
	//based as directly as possible on Oolite's player shield recharge code
	if (delta == undefined) delta = 1;
	//if (this.ship == player.ship.target) log (this.name, "Fore/Aft updating recharge for "+this.ship.name+" delta: "+ delta);
	this.NShields_lastRechargeTime = clock.absoluteSeconds;
	if (debugNeeded) log (this.name, "last recharge time: "+this.NShields_lastRechargeTime+ isNaN(this.NShields_lastRechargeTime))
	
	var fwdMax = this.ship.maxForwardShield;
	var aftMax = this.ship.maxAftShield;
	var shieldRechargeFwd = this.ship.forwardShieldRechargeRate * delta;
	var shieldRechargeAft = this.ship.aftShieldRechargeRate * delta;
	if (debugNeeded) log(this.name, "recharge(1): " + shieldRechargeFwd + ", "+shieldRechargeAft)
	
	var rechargeFwd = Math.min(shieldRechargeFwd, fwdMax - this.ship.forwardShield);
	var rechargeAft = Math.min(shieldRechargeAft, aftMax - this.ship.aftShield);
	if (debugNeeded) log(this.name, "recharge(2): " + shieldRechargeFwd + ", "+shieldRechargeAft)
	
	var energyForShields = this.ship.energy;
	if ( (this.ship.forwardShield > fwdMax * 0.25) && (this.ship.aftShield > aftMax * 0.25) )
	{
		energyForShields = Math.max(0.0, this.ship.energy - (this.ship.maxEnergy * worldScripts.NShields.$shield_charge_energybank_threshold));
	}
	if (debugNeeded) log(this.name, "recharge(3): " + shieldRechargeFwd + ", "+shieldRechargeAft)
	
	if (this.ship.forwardShield < this.ship.aftShield)
	{
		rechargeFwd = Math.min(rechargeFwd, energyForShields);
		rechargeAft = Math.min(rechargeAft, energyForShields - rechargeFwd);
	}
	else
	{
		rechargeAft = Math.min(rechargeAft, energyForShields);
		rechargeFwd = Math.min(rechargeFwd, energyForShields - rechargeAft);
	}
	if (debugNeeded) log(this.name, "recharge(4): " + shieldRechargeFwd + ", "+shieldRechargeAft)
	
	//if (this.ship == player.ship.target) log(this.name, "recharge calculated: "+ rechargeFwd + ", "+rechargeAft);
	this.ship.forwardShield += rechargeFwd;
	this.ship.aftShield += rechargeAft;
	this.ship.energy -= (rechargeFwd + rechargeAft);
	if (debugNeeded) log(this.name, "Shields: "+this.ship.forwardShield +", "+this.ship.aftShield); 
	// Oolite's source code does something called "OOClamp_0_max_f" on the shield and energy values here. I have not included any analogous code because I don't know what the aforementioned function actually does. Hopefully this will not break things.
	
	if (this.ship.forwardShield == fwdMax && this.ship.aftShield == aftMax) this.NShields_rechargeTimer.stop();
}

this.$resetRechargeTimer = function()
{
	this.NShields_rechargeTimer.stop()
	this.NShields_rechargeTimer.nextTime = clock.absoluteSeconds + 1;
	this.NShields_rechargeTimer.start()
}

this.$applyHeatDamage = function(ship, amount)
{
	//log(this.name, "heat dmg for ship: "+ship.name)
	var n_shields = ship.script.NShields_labels.shields.length;
	var pass = 0; //amount passed on to energy banks
	
	for (S in ship.script.NShields_labels.shields)
	{
		var shield_amt = amount / n_shields;
		ship[ship.script.NShields_labels.shields[S]] -= shield_amt;
		if (ship[ship.script.NShields_labels.shields[S]] < 0)
		{
			pass -= ship[ship.script.NShields_labels.shields[S]];
			ship[ship.script.NShields_labels.shields[S]] = 0;
		}
	}
	//log (this.name, "passing damage to energy banks: "+pass)
	ship.energy -= pass;
	//log(this.name, "energy remaining: "+ship.energy)
	this.$applyDamageHandlers(ship, pass, null, "heat damage")
}

this.$applyDamageHandlers = function(ship, amount, whom, type)
{
	//amount -= shieldLevel;
	//if (amount < 0) amount = 0;
	for (f in ship.script.NShields_damageHandlers)
	{
		ship.script.NShields_damageHandlers[f].apply(ship.script, [amount, whom, type]); // check this works
	}
}

this.$drawShieldFX = function (position, shieldValue, ship, attacker, shieldLabel)
{
	if (shieldValue < 1) return; // shields effectively down, no draw
	
	var colour_idx = Math.min(Math.floor(shieldValue/32), (ship.script.NShields_flareColours.length - 1));  //find index for colour array
	//if (colour_idx >= ship.script.NShields_flareColours.length) colour_idx = ship.script.NShields_flareColours.length - 1;
	//if (ship == player.ship.target) log (this.name, "index: "+colour_idx + " length "+ ship.script.NShields_flareColours.length);
	
	if (this.$drawFlashers)
	{
		var shieldFlare = ship.spawnOne("NShields_shieldFlare");
		shieldFlare.position = position;
		if (attacker) shieldFlare.orientation = new Vector3D().rotationTo(attacker.position.subtract(ship.position));//attacker.position.subtract(ship.position).rotationTo([0,0,1]); //attacker may be null if cloaked
		shieldFlare.addCollisionException(ship);
		var scaleFactor = (ship.collisionRadius + Math.min(ship.collisionRadius, (ship.collisionRadius * 0.1 * colour_idx))) * this.$flasherScale;
		for (f in shieldFlare.flashers)
		{
			shieldFlare.flashers[f].color = ship.script.NShields_flareColours[colour_idx];
			shieldFlare.flashers[f].size *= scaleFactor;
			shieldFlare.flashers[f].position = shieldFlare.flashers[f].position.multiply(scaleFactor); //need orientation setting for this to be useful
		}
		shieldFlare.script.$parent = ship;
		shieldFlare.velocity = ship.velocity;
	}
	
	if (this.$drawExploders)
	{
		var expl_idx = Math.min(colour_idx, (ship.script.NShields_exploders.length - 1));
		var expltype = ship.script.NShields_exploders[expl_idx] + ship.script.NShields_sizeClass;
		var shieldExpl = ship.spawnOne(expltype);
		if (shieldFlare) shieldExpl.addCollisionException(shieldFlare);
		shieldExpl.position = position;
		shieldExpl.explode();
	}
	
	if (this.$drawHemiGlow)
	{
		if (shieldLabel == "forwardShield")
		{
			if (ship.script.NShields_hemi_fwd) ship.script.NShields_hemi_fwd.script.$reset(); // || !ship.script.NShields_hemi_fwd.isValid)
			else
			{
				log (this.name, "spawning forward shield effect for "+ship.name)
				ship.script.NShields_hemi_fwd = system.addVisualEffect("N-Shields-hemisphereGlow", ship.position);
				ship.script.NShields_hemi_fwd.scale(ship.collisionRadius * 2);
				ship.script.NShields_hemi_fwd.script.$parent = ship;
				ship.script.NShields_hemi_fwd.script.$label = "NShields_hemi_fwd";
			}
			
			ship.script.NShields_hemi_fwd.setMaterials({None:{emission_color:ship.script.NShields_flareColours[colour_idx]}});
			ship.script.NShields_hemi_fwd.orientation = ship.orientation;
		}
		else if (shieldLabel == "aftShield")
		{
			if (ship.script.NShields_hemi_aft) ship.script.NShields_hemi_aft.script.$reset();// || !ship.script.NShields_hemi_aft.isValid)
			else
			{
				log (this.name, "spawning aft shield effect for "+ship.name)
				ship.script.NShields_hemi_aft = system.addVisualEffect("N-Shields-hemisphereGlow", ship.position);
				ship.script.NShields_hemi_aft.scale(ship.collisionRadius * 2);
				ship.script.NShields_hemi_aft.script.$parent = ship;
				ship.script.NShields_hemi_aft.script.$label = "NShields_hemi_aft";
			}
			
			ship.script.NShields_hemi_aft.setMaterials({None:{emission_color:ship.script.NShields_flareColours[colour_idx]}});
			ship.script.NShields_hemi_aft.orientation = new Quaternion([0, 0, 1, 0]).multiply(ship.orientation);
		}
		else log(this.name, "WARNING: shield label not found: "+shieldLabel + " for ship "+ship.name);
	}
}

this.$drawSparks = function(position, ship, amount) //TODO: use amount to determine type?
{
	if (amount < 10) var chance = 0.5;
	else var chance = 1;
	
	if (Math.random() <= chance)
	{
		var sparks = ship.spawnOne("NShields-sparks");
		sparks.position = position;
		sparks.explode()
	}
}

// Handler for NPC Equipment Damage, called with 'this' referring to ship script

this.NPC_shield_equipmentDamaged = function(equipment)
{
	var shipconf = worldScripts.ShipConfiguration_Core
	if (shipconf)
	{
		if (shipconf._frontshields.indexOf(equipment) >= 0 || shipconf._aftshields.indexOf(equipment) >= 0)
		{
			// Damage shield boosters etc. first if available
			if (this.ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") == "EQUIPMENT_OK")
			{
				this.ship.setEquipmentStatus(equipment, "EQUIPMENT_OK")
				this.ship.setEquipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER", "EQUIPMENT_DAMAGED")
				equipment = "EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER"; // so later code will immediately deactivate the damaged booster
			}
			else if (this.ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") == "EQUIPMENT_OK")
			{
				this.ship.setEquipmentStatus(equipment, "EQUIPMENT_OK")
				this.ship.setEquipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER", "EQUIPMENT_DAMAGED")
				equipment = "EQ_NSHIELDS_NPC_SHIELD_BOOSTER";
			}
			else // actually damage the basic shield generator
			{
				if (shipconf._frontshields.indexOf(equipment) >= 0) this.ship.forwardShieldRechargeRate = -0.5;
				if (shipconf._aftshields.indexOf(equipment) >= 0) this.ship.aftShieldRechargeRate = -0.5;
			}
		}
	}
	
	if (equipment == "EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") //reverse the equipment upgrade changes
	{
		let eq = EquipmentInfo.infoForKey("EQ_NAVAL_SHIELD_BOOSTER");
		let shield_boost = parseInt(eq.scriptInfo.oolite_shield_increase);
		let recharge_boost = parseFloat(eq.scriptInfo.oolite_shield_recharge_multiplier);
		for (i in this.NShields_labels.maxShields)
		{
			this.ship[this.NShields_labels.maxShields[i]] -= shield_boost;
			if (recharge_boost) this.ship[this.NShields_labels.rechargeRates[i]] /= recharge_boost;
		}
	}
	
	if (equipment == "EQ_NSHIELDS_NPC_SHIELD_BOOSTER") //reverse the equipment upgrade changes
	{
		let eq = EquipmentInfo.infoForKey("EQ_SHIELD_BOOSTER");
		let shield_boost = parseInt(eq.scriptInfo.oolite_shield_increase);
		let recharge_boost = parseFloat(eq.scriptInfo.oolite_shield_recharge_multiplier);
		for (i in this.NShields_labels.maxShields)
		{
			this.ship[this.NShields_labels.maxShields[i]] -= shield_boost;
			if (recharge_boost) this.ship[this.NShields_labels.rechargeRates[i]] /= recharge_boost;
		}
	}
}

// shipTakingDamage handler component for NPC Shield Cyclers

this.NPC_sc_adjust = function(amount, whom, type)
{
	//check cycler is active 1st?
	var adjust_input = 
	{
		caller: "N-Shields",
		setting: 3, //equal. TODO: choose appropriate setting somewhere
		functional: 100, //TODO: damage % via NPC Equipment Damage OXP?
		fwd: this.ship.forwardShield,
		fwd_threshold: (this.ship.maxForwardShield * worldScripts["Shield Cycler"]._sc_settings.sc.threshold * 0.01 ),
		aft: this.ship.aftShield,
		aft_threshold: (this.ship.maxAftShield * worldScripts["Shield Cycler"]._sc_settings.sc.threshold * 0.01 ),
		version: this.NShields_SC_version,
		manual_version: this.NShields_SC_manual_version,
		init: false,
		max_energy: this.ship.maxEnergy
	};
	
	var adjust_output = worldScripts["SC Core"]._sc_adjust(adjust_input);
	
	this.ship.forwardShield = Math.min(adjust_output.fwd, this.ship.maxForwardShield);
	this.ship.aftShield = Math.min(adjust_output.fwd, this.ship.maxAftShield);
	this.ship.energy -= adjust_output.power;
}